"""Basic datatypes for the superspy language.
"""

from superspy import ast, language


@language.register_data_type(r'^(\d*\.)?\d+$')
class Number(ast.DataType):
    """A number that takes floats and ints and stores them as float.
    """

    def __init__(self, line_number: int, content: str):
        """Initialize and set up the number from a string from the tree.
        """
        super().__init__(line_number, content)
        self.value = float(content)


@language.register_literality_delimiter('"', '"')
@language.register_data_type(r'^".*"$')
class String(ast.DataType):
    """A string.
    """

    def __init__(self, line_number: int, content: str):
        """Initialize and set up the string from the tree.
        """
        super().__init__(line_number, content)
        self.value = content[1:-1]



"""Manage the setting of variables in the superspy language.
"""

from superspy import ast, language


@language.register_token('=')
class SetSymbol(ast.Token):
    """A `=` token.
    """
    pass


@language.register_command
class SetVariable(ast.Command):
    """Set a variable.

    Attributes:
        value (ast.Command): The value to set the variable to.
        variable (ast.Variable): The variable to set.
    """

    priority = ast.OrderOfOperations.ASSIGNMENT

    variable: ast.Variable
    value: ast.Command

    def __init__(self,
                 variable: ast.Variable,
                 _set_symbol: SetSymbol,
                 value: ast.Command):
        """Initialize the set command.

        Args:
            variable (ast.Variable): The variable to set.
            _set_symbol (SetSymbol): The `=` token.
            value (ast.Command): The value to set the variable to.
        """
        super().__init__()
        self.variable = variable
        self.value = value

    def execute(self):
        """Execute the value and set the variable to it.
        """
        self.variable.set(self.value.execute())

    def __repr__(self) -> str:
        return f'SET {self.variable.content} TO ({repr(self.value)})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}SET\n'\
            f'{self.variable.deep_repr(indent+self.addint)}\n'\
            f'{indent+self.addint}=\n'\
            f'{self.value.deep_repr(indent+self.addint)}'



"""Basic flow control functions within the superspy language.
"""

from typing import Optional

from superspy import ast, language
from superspy.system import base_language


@language.register_token('if')
class IfToken(ast.Token):
    """An `if` token.
    """
    pass

@language.register_token('else')
class ElseToken(ast.Token):
    """An `else` token.
    """
    pass


@language.register_falsy_value(0)
@language.register_falsy_value(0.0)
@language.register_falsy_value('0')
@language.register_command
class IfCondition(ast.Command):
    """A simple if condition without an else case.

    Attributes:
        condition (ast.Command): The condition determining the execution.
        group (ast.Command): The code to be executed if the condition holds.
    """

    priority = ast.OrderOfOperations.FLOW_CONTROL

    condition: ast.Command
    group: ast.Command

    def __init__(self,
                 _iftoken: IfToken,
                 condition: ast.Command,
                 _newline: Optional[base_language.NewLine],
                 group: ast.Command):
        """Initialize the if condition.

        Args:
            _iftoken (IfToken): The `if` token.
            condition (ast.Command): The condition determining the execution.
            _newline (Optional[base_language.NewLine]): An optional newline
                character after the condition before the opening brace or other
                command to be executed.
            group (ast.Command): The code to be executed if the condition
                holds true.
        """
        super().__init__()
        self.condition = condition
        self.group = group

    def execute(self):
        """Execute the statement, its condition, and possibly its group.
        """
        if self.ast.runtime.language.evaluates_as_true(
                self.condition.execute()):
            self.group.execute()

    def __repr__(self) -> str:
        return f'CONDITION ({repr(self.condition)}) '\
            f'IF TRUE ({repr(self.group)})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}CONDITION\n'\
            f'{self.condition.deep_repr(indent+self.addint)}\n'\
            f'{indent}IF TRUE\n'\
            f'{self.group.deep_repr(indent + self.addint)}'


@language.register_command
class IfElseCondition(ast.Command):
    """A simple if condition with an else case.

    Attributes:
        condition (ast.Command): The condition determining the execution.
        group_true (ast.Command): The code to be executed if the condition
            yields True.
        group_false (ast.Command): The code to be executed if the condition
            yields False.
    """

    priority = ast.OrderOfOperations.FLOW_CONTROL - 0.1

    condition: ast.Command
    group_true: ast.Command
    group_false: ast.Command

    def __init__(self, _iftoken: IfToken,
                 condition: ast.Command,
                 _newline1: Optional[base_language.NewLine],
                 group_true: ast.Command,
                 _newline2: Optional[base_language.NewLine],
                 _elsetoken: ElseToken,
                 _newline3: Optional[base_language.NewLine],
                 group_false: ast.Command):
        """Initialize the if/else condition.

        Args:
            _iftoken (IfToken): The `if` token.
            condition (ast.Command): The codition determining the execution.
            _newline1 (Optional[base_language.NewLine]): An optional newline.
            group_true (ast.Command): The code to be executed if the condition
                yields True.
            _newline2 (Optional[base_language.NewLine]): An optional newline.
            _elsetoken (ElseToken): The `else` token.
            _newline3 (Optional[base_language.NewLine]): An optional newline.
            group_false (ast.Command): The code to be executed if the condition
                yields False.
        """
        super().__init__()
        self.condition = condition
        self.group_true = group_true
        self.group_false = group_false

    def execute(self):
        """Execute the statement by executing its condtion and then one group.
        """
        if self.ast.runtime.language.evaluates_as_true(
                self.condition.execute()):
            self.group_true.execute()
        else:
            self.group_false.execute()

    def __repr__(self) -> str:
        return f'CONDITION ({repr(self.condition)}) '\
            f'IF TRUE ({repr(self.group_true)}) '\
            f'IF FALSE ({repr(self.group_false)})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}CONDITION\n'\
            f'{self.condition.deep_repr(indent+self.addint)}\n'\
            f'{indent}IF TRUE\n'\
            f'{self.group_true.deep_repr(indent + self.addint)}\n'\
            f'{indent}IF FALSE\n'\
            f'{self.group_false.deep_repr(indent + self.addint)}'


@language.register_token('while')
class WhileToken(ast.Token):
    """A `while` token.
    """
    pass


@language.register_command
class WhileLoop(ast.Command):
    """A while loop repeatedly executing a command while a condition is True.

    Attributes:
        condition (ast.Command): The codition determining the execution.
        group (ast.Command): The group to execute.
    """

    priority = ast.OrderOfOperations.FLOW_CONTROL

    condition: ast.Command
    group: ast.Command

    def __init__(self,
                 _whiletoken: WhileToken,
                 condition: ast.Command,
                 _newline: Optional[base_language.NewLine],
                 group: ast.Command):
        """Initialize the while loop.

        Args:
            _whiletoken (WhileToken): The `while` token.
            condition (ast.Command): The condition determining the execution.
            _newline (Optional[base_language.NewLine]): An optional newline.
            group (ast.Command): The group to execute while the condition is
                True.
        """
        super().__init__()
        self.group = group
        self.condition = condition

    def execute(self):
        """Check the condition and the running state of the ast for running.
        """
        while self.condition.executes_as_true() and self.ast.is_running():
            self.group.execute()

    def __repr__(self) -> str:
        return f'LOOP WHILE ({repr(self.condition)}): ({repr(self.group)})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}LOOP WHILE\n'\
            f'{self.condition.deep_repr(indent+self.addint)}\n'\
            f'{self.group.deep_repr(indent + self.addint)}'


@language.register_token('exit')
class ExitToken(ast.Token):
    """An `exit` token.
    """
    pass


@language.register_command
class Exit(ast.Command):
    """Exit the program with an optional exit code.

    Attributes:
        exit_code (ast.Command): The exit code to end with.
    """

    priority = ast.OrderOfOperations.EXPRESSION

    exit_code: Optional[ast.Command]

    def __init__(self, _exittoken: ExitToken,
                 exit_code: Optional[ast.Command]):
        """Summary

        Args:
            _exittoken (ExitToken): The exit token.
            exit_code (Optional[ast.Command]): An optional exit code.
        """
        super().__init__()
        self.exit_code = exit_code

    def execute(self):
        """Set the exit code of the ast depending on the argument.
        """
        exit_code = 0
        if self.exit_code is not None:
            exit_code = self.exit_code.execute()
        self.ast.runtime.exit_code = exit_code

    def __repr__(self) -> str:
        if self.exit_code is not None:
            return f'EXIT ({repr(self.exit_code)})'
        return f'EXIT'

    def deep_repr(self, indent='') -> str:
        if self.exit_code is not None:
            return f'{indent}EXIT\n'\
            f'{self.exit_code.deep_repr(self.addint + indent)}'
        return f'{indent}EXIT'



"""This is the core of superspy; it handles most of the parsing and execution.
"""

import re
from abc import ABC, abstractmethod

from enum import Enum, auto
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from superspy import code_source, language


# pylint: disable=too-few-public-methods
# This class is not supposed to have methods, it is a bit like an enum.
class OrderOfOperations:
    """Order of operations priorities reference.
    They are floats, so values in-between can be chosen if so desired.
    The default library limits itself to those values, some commands are
    + or - 0.1 if they need to be parsed before or after a specific other
    command.

    Adapted from Wikipedia: https://en.wikipedia.org/wiki/Order_of_operations
    """
    FUNCTION_CALL = 1.0
    UNARY = 2.0
    MULTIPLICATION = 3.0
    ADDITION = 4.0
    BIT_SHIFT = 5.0
    UNEQUAL_COMPARISON = 6.0
    EQUAL_COMPARISON = 7.0
    AND_BW = 8.0
    XOR_BW = 9.0
    OR_BW = 10.0
    AND_LOGICAL = 11.0
    OR_LOGICAL = 12.0
    TERNARY = 13.0
    ASSIGNMENT = 14.0
    EXPRESSION = 15.0
    BRACES = 16.0
    FLOW_CONTROL = 17.0
    COMMA = 18.0
    UNDEFINED = 100.0
    END = 10000.0


class ErrorMode(Enum):
    """The modes an interpreter can have handling errors.

    Attributes:
        CRASH (TYPE): Crash once an error is encountered. This is the default
            in general, but should be especially for any execution of pre-made
            code, since correct execution cannot be guaranteed once one line of
            code fails.
        PRINT (TYPE): This one prints the error, but continues execution. It is
            used in the shell, where you don't want the shell to end when you
            accidentally type a variable name wrong.
        SUPPRESS (TYPE): Ignore the error silently and continue executing the
            code. This one is not used in here anywhere, but can be used by
            you. Beware however, since this might lead to issues!
        MIRROR_PARENT (TYPE): Mirror the parent ast's error mode.
    """

    SUPPRESS = auto()
    PRINT = auto()
    CRASH = auto()
    MIRROR_PARENT = auto()


###########
# CLASSES #
###########


class Token:
    """The base class of any token in the entire project.

    Tokens can be single characters like `\n`, `*` or `(`, words like `print`,
    strings, numbers, or entire statements with multiple children, so for
    example `print 4 + 3` would in the end be a Token subclass for the print
    command, with a child being the addition, itself having the children `4`,
    `3` and `+`.

    Attributes:
        line_number (int): The line number this token ends at.
        content (str): The string content this token describes.

        ast (Ast): The AST object managing this token.

        parent (Optional[Token]): The parent object.
        children (List[Token]): The children objects.

        has_built_token_tree (bool): Has this token already built a token tree.

        addint (str): Additional indent for pretty print formatting.
    """

    line_number: int
    content: str
    ast: Optional['Ast'] = None
    addint = '  '
    parent: Optional['Token'] = None
    children: List['Token'] = []
    has_built_token_tree = False

    def __init__(self, line_number: Optional[int] = None,
                 content: Optional[str] = None):
        """Initiate the token

        Args:
            line_number (int, optional): The line number this token ends at.
            content (str, optional): The string content this token describes.
        """
        if line_number is not None:
            self.line_number = line_number
        if content is not None:
            self.content = content

    def __repr__(self) -> str:
        """String representation of this object.
        """
        return f'{type(self).__name__} '\
               f'at line {self.line_number}: {repr(self.content)}'

    def deep_repr(self, indent='') -> str:
        """A deep and indented string representation of the object.

        Args:
            indent (str, optional): The indent to add before this token.

        Returns:
            str: The string representation.
        """
        return f'{indent}{type(self).__name__}[{self.line_number}]'

    def namespace(self) -> str:
        """The namespace for this object and its children.
        So for example a variable inside a loop (like i) would have a different
        namespace than a variable outside to loop, so it cannot be accessed
        there.

        Returns:
            str: The namespace.
        """
        if self.parent is None:
            return ''
        return self.parent.namespace()

    def build_token_tree(self):
        """Build the token tree of all the children.
        This is mostly only relevant for groups like braces which still have to
        be built after being grouped.
        """
        if not self.has_built_token_tree:
            self.has_built_token_tree = True
            for child in self.children:
                child.build_token_tree()

    def get_trace(self) -> str:
        """Get's a trace for this Token which can be used in error messages.

        Returns:
            str: the trace.
        """
        if self.parent is None:
            return ''
        return f'{self.parent.get_trace()}\n{repr(self)}'

    # pylint: disable=no-self-use
    # This is a method supposed to be inherited.
    def executes_as_true(self) -> bool:
        """Is this token True or False.

        Returns:
            bool: The boolean value of this token.
        """
        return False


class UndefinedToken(Token):
    """A not defined token.
    """
    pass


class Command(Token, ABC):
    """A command that can be made up of different tokens.
    It is described by a pattern made up of keywords and arguments in any
    order. It even allows for lists.

    Attributes:
        priority (float): The priority to be parsed in. Should be derived from
            OrderOfOperations. Smaller numbers get matched first.
        pattern (List[language.PATTERN]): A pattern to be matched. Look in the
            standard library for reference.
        can_contain_self (bool): Can the command contain itself.
    """

    priority: float = OrderOfOperations.UNDEFINED
    pattern: List[language.PATTERN] = None
    can_contain_self = True

    @abstractmethod
    def execute(self) -> Any:
        """Execute this command.
        """
        pass

    def executes_as_true(self) -> bool:
        """Is the value when executing this command Truthy or Falsy.

        Returns:
            bool: The return value
        """
        return self.execute() not in self.ast.runtime.language.falsy_values

    @classmethod
    def token_matches_class(cls, token: Token, class_: Type[Token]) -> bool:
        """Does this token conform to a class.

        Args:
            token (Token): The token to match.
            class_ (Type[Token]): The class to match to.

        Returns:
            bool: The return value.
        """
        if token is None or class_ is None:
            return False
        return isinstance(token, class_)

    # pylint: disable=too-many-arguments
    # This many arguments are unfortunately required.
    @classmethod
    def match_single_token(cls,
                           token_list: List[Token],
                           pattern: Type[Token],
                           next_pattern: Type[Token],
                           is_optional: bool,
                           is_list: bool) -> int:
        """Does a single pattern match one or more tokens.

        Args:
            token_list (List[Token]): A list of tokens to be matched by the
                pattern.
            pattern (Type[Token]): The pattern the token[s] have to conform to.
            next_pattern (Type[Token]): The pattern the next token has to
                conform to. Used for optionals, which can return 0 if the next
                token matches, and lists, which will match as many tokens as
                possible until the next matches.
            is_optional (bool): Is this token optional.
            is_list (bool): Is this token a list, so it matches multiple.

        Returns:
            int: The amounts of tokens matched by this pattern. -1 means it did
                not match successfully, while 0 means it matches 0 tokens,
                which is possible with optionals.
        """

        # Amount of tokens to match.
        length = 0

        while 1:
            # Reached the end.
            if len(token_list) == length:
                if length == 0 and not is_optional:
                    length = -1
                break
            next_token = token_list[length]
            if (not cls.can_contain_self) and\
                    cls.token_matches_class(next_token, cls):
                length = -1
                break

            # Next pattern matches, which limits the match.
            if cls.token_matches_class(next_token, next_pattern) and (
                    True if is_optional else length > 0):
                break

            # Matches current pattern.
            if cls.token_matches_class(next_token, pattern):
                length += 1
            # Does not match the current pattern.
            else:
                # Optionals still are successful.
                if is_optional:
                    break
                length = -1
                break

            # Only lists match multiple tokens.
            if not is_list:
                break

        return length

    @classmethod
    def type_of_token(cls, token: language.PATTERN
                      ) -> Tuple[Type[Token], bool, bool]:
        """Returns the type of a single token.
        It strips the `List` and `Optional` attributes off and returns them as
        booleans.

        Args:
            token (language.PATTERN): The token to analyze.

        Returns:
            Tuple[
                Type[Token],: The type of the token.
                bool,       : Is the token a list.
                bool        : Is the token optional.
            ]
        """
        is_list = False
        is_optional = False

        token_type = token

        # This is a bit hacky, but origin contains information about things
        # like List or Optionality for `List[Token]` or `Optional[Token]`.
        origin = token.__origin__ if hasattr(token, '__origin__') else None
        # The arguments contains information about the type of object.
        args = token.__args__ if hasattr(token, '__args__') else None

        # Matches list.
        if origin is list:
            token_type = args[0]
            is_list = True
        if origin is Union:
            # Defines NoneType, which cannot be accessed otherwise.
            NoneType = type(None)
            # An optional is a union with NoneType.
            if NoneType in args:
                is_optional = True
                # The other argument is the type of token.
                if args[0] is NoneType:
                    token_type = args[1]
                else:
                    token_type = args[0]

        return token_type, is_list, is_optional

    @classmethod
    def match(cls, token_list: List[Token]) -> List[int]:
        """Match a list of tokens to this classes pattern.

        Args:
            token_list (List[Token]): The list of tokens to match.

        Returns:
            List[int]: The amount of tokens matched by each element of the
                pattern.
        """
        length = 0
        return_list: List[int] = []

        for pattern_index, pattern_token_class in enumerate(cls.pattern):
            current_token_type, is_list, is_optional = cls.type_of_token(
                pattern_token_class)
            next_token_type = None
            if len(cls.pattern) > pattern_index + 1:
                # Get the type of the next token if there is one.
                next_token_type, _, _ = cls.type_of_token(
                    cls.pattern[pattern_index + 1])

            # Reached the end without matching the entire pattern.
            if len(token_list) == length and not is_optional:
                return []

            new_length = cls.match_single_token(
                token_list[length:],
                current_token_type,
                next_token_type,
                is_optional,
                is_list)
            # Did not match sucessfully
            if new_length == -1:
                return []
            return_list.append(new_length)
            length += new_length

        if [element for element in return_list if element > 0]:
            # Matched not only empty optionals.
            return return_list
        return []


class Function(Command, ABC):
    """A function which is a command beginning with one keyword.

    Attributes:
        args (List[Token]): The argument tokens for this instance.
        argument (Token): The first argument. Used when there only is one.
    """

    args: List[Token] = None
    argument: Token = None
    priority: float = OrderOfOperations.EXPRESSION

    def __init__(self, _: Token, *args: List[Token]):
        """Set up from the default initializer.

        Args:
            _ (Token): The function name.
            *args (List[Token]): The arguments
        """
        super().__init__()
        self.args = args
        if args:
            self.argument = args[0]

    @abstractmethod
    def execute(self) -> Any:
        """Execute this function. Overwriting it like this is required
        by pylint.
        """
        pass


class Operation(Command, ABC):
    """An operation which has an operator in the middle (like `+`) and an
    argument on each side.

    Attributes:
        left (Command): The left parameter of the operation.
        right (Command): The right parameter of the operation.
    """

    left: Command
    right: Command

    def __init__(self, left: Command, _: Token, right: Command):
        """Set up from the default initializer.

        Args:
            left (Command): The left parameter of the operation.
            _ (Token): The operator.
            right (Command): The right parameter of the operation.
        """
        super().__init__()
        self.left = left
        self.right = right

    @abstractmethod
    def execute(self) -> Any:
        """Execute this function. Overwriting it like this is required
        by pylint.
        """
        pass


class DataType(Command, ABC):
    """A datatype / value.
    """

    value: Any

    def execute(self) -> Any:
        """Return the value.

        Returns:
            Any: The value.
        """
        return self.value

    def __repr__(self) -> str:
        return f'{type(self).__name__} {repr(self.value)}'

    def deep_repr(self, indent='') -> str:
        return f'{indent}{type(self).__name__}={repr(self.value)}'


class Variable(Command):
    """A variable that support setting and getting a value.

    Attributes:
        name (str): The name of the variable.
    """

    name: str

    def __init__(self, name: str):
        """Initialize the object and set its name

        Args:
            name (str): The name of the variable.
        """
        super().__init__()
        self.name = name

    def get(self) -> Any:
        """Get the variable's value.

        Returns:
            Any: The variable's value.
        """
        search_namespace = ''
        # Iterate through all the parent namespaces and check if a variable
        # with that name exists there somewhere.
        for namespace_component in self.namespace().split('/'):
            search_namespace = f'{search_namespace}/{namespace_component}'
            variable_name = f'{search_namespace}.{self.name}'
            # If that variable is found return it.
            if variable_name in self.ast.runtime.variables:
                return self.ast.runtime.variables[variable_name]
        # If the variable is not found throw an error.
        self.ast.error(f'Variable or function {self.name} not found.',
                       self.get_trace())
        return None

    def set(self, new_value: Any):
        """Set the variable's value.

        Args:
            new_value (Any): The variable's new value.
        """
        search_namespace = ''
        # Iterate through all the parent namespaces and check if a variable
        # with that name exists there somewhere.
        for namespace_component in self.namespace().split('/'):
            search_namespace = f'{search_namespace}/{namespace_component}'
            variable_name = f'{search_namespace}.{self.name}'
            # If the variable is found set its value.
            if variable_name in self.ast.runtime.variables:
                self.ast.runtime.variables[variable_name] = new_value
                break
        else:
            # If the variable is not found set a new one.
            self.ast.runtime.variables[
                f'{search_namespace}.{self.name}'] = new_value

    def execute(self) -> Any:
        """If the variable gets executed like a command its value is returned.

        Returns:
            Any: The variable's value.
        """
        return self.get()

    def __repr__(self) -> str:
        return f'VARIABLE {self.name}'

    def deep_repr(self, indent='') -> str:
        return f'{indent}VARIABLE {self.name}'


class Runtime:
    """Class that stores runtime data.
    It has a single instance for a running script.

    Attributes:
        root_ast (Ast): The root ast of the script.
        exit_code (optional, int): The exit code of the script.
        language (language.Language): The language of the script.
        variables (Dict[str, Any]): The variables in the script.
        functions (Dict[str, Any]): The functions in the script.
    """

    root_ast: 'Ast'
    language: language.Language
    exit_code: Optional[int] = None
    variables: Dict[str, Any] = {}
    functions: Dict[str, Any] = {}

    def __init__(self, root_ast: 'Ast'):
        """Initialize and set up the Runtime.

        Args:
            root_ast (Ast): The root ast of the script.
        """
        self.root_ast = root_ast


class Ast:
    """The Abstract Syntax Tree doing most of the heavy lifting of the parsing.
    This manages all the parsing and executing of the input.

    Attributes:
        ast (List[Token]): The actual abstract syntax tree: A list of tokens.
            This is a flat list, because any depth are the sub tokens of
            another token.
        ast_is_definitely_valid_up_to (int): When parsing the AST from a
            continuing input like a shell the AST tries to match at every line.
            However with asymmetric braces it might not yet be valid. This is
            the last point the AST is entirely valid at, so the previous part
            does not have to be parsed again.
        lang (language.Language): The language this ast is parsing, matching
            and executing.
        line_number (int): The current line number the ast is parsing at.
        error_mode (ErrorMode): What to do when an error occurs, like crash or
            continue executing.
        source (code_source.CodeSource): The code source for this ast that
            supplies new lines of code.
        execution_index (int): The index the execution is currently at. This
            has to be a nonlocal variable because the shell terminates the
            execution function between lines.
    """

    ast: List[Token] = []
    source: Optional[code_source.CodeSource] = None
    error_mode: ErrorMode = ErrorMode.CRASH

    ast_is_definitely_valid_up_to = 0
    line_number = 0

    execution_index: int = 0

    runtime: Optional[Runtime] = None

    def __init__(self, source: Optional[code_source.CodeSource] = None,
                 lang: Optional[language.Language] = None):
        """Initialize the current AST with a code source and a language.

        Args:
            source (Optional[code_source.CodeSource]): The code source this ast
                draws from.
            lang (language.Language): The language this ast matches.
        """
        self.source = source
        if self.runtime is None:
            self.runtime = Runtime(self)
            if language is not None:
                self.runtime.language = lang

    def build_token_list(self, line_count: int = None):
        """Build a flat token list from strings by the code source.
        New lines are requested, either until it reaches the end or a specified
        limit. It then parses it character by character and matches it to
        elementary tokens to build a flat token list.

        Args:
            line_count (int, optional): The amount of lines to parse. If left
                empty or `None` it will run to the end of the source.
        """
        # tree building components
        current_word = ''

        # state variables
        escape_next = False
        is_literal = False
        literality_closer: chr = ''

        def save_current_word():
            """Save the current word to the token list.
            """
            nonlocal current_word
            if current_word != '':
                match_class = UndefinedToken
                # Try to match it to existing tokens.
                for regex in self.runtime.language.regex_matchers:
                    token_class = self.runtime.language.regex_matchers[regex]
                    if re.match(regex, current_word):
                        match_class = token_class
                        break
                # Initialize a new token and set it up.
                new_token = match_class(self.line_number, current_word)
                new_token.ast = self
                # Add it to the token list.
                self.ast.append(new_token)
                # Reset the current word.
                current_word = ''

        # Repeat until end of source or for specified amount of lines.
        while self.source.has_more_lines() and (
                line_count is None or line_count):
            if line_count:
                line_count -= 1
            new_line = self.source.next_line(self.is_valid())
            self.line_number += 1
            for character in new_line:
                if escape_next:
                    current_word += character
                    escape_next = False
                elif character in self.runtime.language.escape_markers:
                    escape_next = True
                elif is_literal and character == literality_closer:
                    is_literal = False
                    current_word += character
                    literality_closer = ''
                elif is_literal:
                    current_word += character
                elif character in \
                        self.runtime.language.literality_delimiter.keys():
                    literality_closer = \
                            self.runtime.language.literality_delimiter[
                                character]
                    is_literal = True
                    current_word += character
                elif character in self.runtime.language.word_delimiters:
                    save_current_word()
                elif character in \
                        self.runtime.language.single_character_tokens:
                    save_current_word()
                    current_word += character
                    save_current_word()
                else:
                    current_word += character
            save_current_word()

    def guess_variables(self):
        """Guess that all remaining tokens are variables.
        """
        for index, token in enumerate(self.ast):
            if isinstance(token, UndefinedToken):
                # Initialize and setup new variable.
                new_var = Variable(token.content)
                new_var.line_number = token.line_number
                new_var.content = token.content
                new_var.ast = self
                self.ast[index] = new_var

    def build_token_tree(self):
        """Match patterns defined for different commands and create them.
        """
        # pylint: disable=too-many-nested-blocks
        # pylint: disable=too-many-locals
        # This is a complex function that belongs together.
        # Spreading it out more would make it less legible.
        for order_of_operation in sorted(
                self.runtime.language.commands_by_priority.keys()):
            # Reverseley walking the list, as to correctly match nested braces
            # and multiple functions.
            current_index = len(self.ast) - 1
            while current_index + 1:
                for command in self.runtime.language.commands_by_priority[
                        order_of_operation]:
                    command.ast = self
                    match = command.match(self.ast[current_index:])
                    breaks_self_matching_rule = False
                    if match and not breaks_self_matching_rule:
                        # Amount of matched
                        match_count = sum(match)

                        remaining_matched_objects = self.ast[
                            current_index:
                            current_index + match_count]
                        flat_matched_objects = remaining_matched_objects.copy()
                        # The command arguments
                        arguments: List[Union[List[Token], Token]] = []
                        for argument_index in range(len(match)):
                            _, is_list, is_optional = command.type_of_token(
                                command.pattern[argument_index])
                            argument_count = match.pop(0)
                            if argument_count == 0 and is_optional:
                                arguments.append(None)
                            elif is_list:
                                # List arguments.
                                arguments_list_arguments: List[Token] = []
                                for _ in range(argument_count):
                                    arguments_list_arguments.append(
                                        remaining_matched_objects.pop(0))
                                arguments.append(arguments_list_arguments)
                            else:
                                # Single argument.
                                arguments.append(
                                    remaining_matched_objects.pop(0))
                        # Create new command object and set it up
                        new_object = command(*arguments)
                        new_object.line_number = arguments[0].line_number
                        new_object.ast = self
                        new_object.children = flat_matched_objects
                        new_object.content = ' '.join(
                            obj.content for obj in flat_matched_objects)
                        for matched_object in flat_matched_objects:
                            matched_object.parent = new_object
                            self.ast.remove(matched_object)
                        self.ast.insert(current_index, new_object)
                        new_object.build_token_tree()
                        # Step "back" one
                        current_index += 1
                current_index -= 1

    def is_valid(self) -> bool:
        """Is the current syntax tree valid.
        This is checked by every token being subclass of command and thus being
        executable.

        Returns:
            bool: The validity.
        """
        for i in range(self.ast_is_definitely_valid_up_to, len(self.ast)):
            if not isinstance(self.ast[i], Command):
                return False
        self.ast_is_definitely_valid_up_to = len(self.ast) - 1
        return True

    def deep_repr(self) -> str:
        """A deep and indented string representation of the syntax tree.

        Returns:
            str: The string representation.
        """
        deep_repr_string = ''.join((c.deep_repr() for c in self.ast))
        while '\n\n' in deep_repr_string:
            deep_repr_string = deep_repr_string.replace('\n\n', '\n')
        return deep_repr_string

    def is_running(self) -> bool:
        """Is the interpreter still running.

        Returns:
            bool: The return value.
        """
        return self.runtime.exit_code is None

    def error(self, message: str, trace: str):
        """Cause an error in the syntax tree that might stop execution.
        The type of reaction depends on the `error_mode`.

        Args:
            message (str): The error message to be displayed.
            trace (str): The trace for the error.
        """
        self.throw_error_with_mode(self.error_mode, message, trace)

    def throw_error_with_mode(self, error_mode: ErrorMode,
                              message: str, trace: str):
        """Throw an error with a specified mode.

        Args:
            error_mode (ErrorMode): The error mode to be executed.
            message (str): The error message to be displayed.
            trace (str): The trace for the error.
        """
        if error_mode == ErrorMode.CRASH:
            print(f'\033[31m{message}\n\n{trace}\033[0m')
            self.runtime.exit_code = 1
        elif error_mode == ErrorMode.PRINT:
            print(f'\033[31m{message}\n\n{trace}\033[0m')
        elif error_mode == ErrorMode.SUPPRESS:
            pass
        elif error_mode == ErrorMode.MIRROR_PARENT:
            print(f'\033[31mAN ERROR OCCURRED,\n'
                  f'BUT ERROR MODE "MIRROR PARENT" IS NOT SUPPORTED\n\n'
                  f'{message}\n\n{trace}\033[0m')

    def interpret(self):
        """Interpret the syntax tree to its end.
        """
        while self.execution_index != len(self.ast) and self.is_running():
            command = self.ast[self.execution_index]
            command.execute()
            self.execution_index += 1

    def complete_run(self, run_after_each_line: bool = False) -> int:
        """Build and run the syntax tree.

        Args:
            run_after_each_line (bool, optional): Run after parsing a single
                line. This is used for the shell mode and is not recommended
                when running entire programs. Otherwise it first parses the
                entire tree and then executes it.

        Returns:
            int: The program exit code.
        """
        while 1:
            self.build_token_list(1 if run_after_each_line else None)
            self.guess_variables()
            self.build_token_tree()
            # print(self.deep_repr())
            if self.is_valid():
                self.interpret()
            elif not run_after_each_line:
                print('INVALID SYNTAX\n\n\n', self.deep_repr())

            if not self.is_running():
                return self.runtime.exit_code or 0

            if not run_after_each_line:
                return 0


class SubAst(Ast):
    """An Ast object that is the child of another Ast.
    This is currently used by the braces/group for managing the Ast contained
    within them.

    Attributes:
        parent_ast (Ast): The parent Ast.
    """

    parent_ast: Ast

    def __init__(self, parent_ast: Ast):
        """Initialize the Ast with a parent.

        Args:
            parent_ast (Ast): The parent Ast.
        """
        self.parent_ast = parent_ast
        self.runtime = self.parent_ast.runtime
        super().__init__()

    def error(self, message: str, trace: str):
        if self.error_mode == ErrorMode.MIRROR_PARENT:
            self.parent_ast.error(message, trace)
        else:
            self.throw_error_with_mode(self.error_mode, message, trace)



"""Basic braces for the superspy language for non flat code.
"""

from typing import List

from superspy import ast, language


@language.register_single_character_touching_token('{')
class OpenBrace(ast.Token):
    """An open brace (`{`) token.
    """
    pass


@language.register_single_character_touching_token('}')
class ClosingBrace(ast.Token):
    """A closing brace (`}`) token.
    """
    pass


@language.register_command
class Group(ast.Command):
    """A group or closure which contains multiple lines of code.
    It groups multiple lines of code into one command, which can be used by
    flow control commands like `if` or `while`.

    Attributes:
        commands (List[ast.Command]): The list of commands inside the group.
        sub_ast (ast.SubAst): The sub ast managing the parsing and interpreting
            of the group.
    """

    commands: List[ast.Command]
    priority = ast.OrderOfOperations.BRACES
    sub_ast: ast.SubAst

    def __init__(self,
                 _ob: OpenBrace,
                 commands: List[ast.Token],
                 _cb: ClosingBrace):
        """Initialize and set up the group

        Args:
            _ob (OpenBrace): The opening brace.
            commands (List[ast.Token]): The commands inside the group.
            _cb (ClosingBrace): The clsing brace.
        """
        super().__init__()
        self.commands = commands

    def execute(self):
        """Interprete the group.
        """
        self.sub_ast.interpret()
        self.sub_ast.execution_index = 0

    def namespace(self):
        """Summary

        Returns:
            TYPE: Description
        """
        return f'{super().namespace()}/GROUP@{self.line_number}'

    def build_token_tree(self):
        """Build the token tree inside the group.
        This is necessary since the group might be created during the parsing
        of the main ast and the commands might not be completeley transformed
        into the final token tree.
        """
        if not self.has_built_token_tree:
            self.has_built_token_tree = True
            self.sub_ast = ast.SubAst(self.ast)
            self.sub_ast.ast = self.commands
            self.sub_ast.build_token_tree()

    def __repr__(self) -> str:
        return f'GROUP {self.line_number}: '\
            f'{" ".join(repr(command) for command in self.commands)}\n'

    def deep_repr(self, indent='') -> str:
        command_string = "".join(command.deep_repr(indent+self.addint)
                                 for command in self.commands)
        return f'{indent}GROUP[{self.line_number}]\n{command_string}\n'



"""Specify languages and register new commands, functions and other tokens.

When creating a new Language object (usually one should be enough) it will
set the Language.current attribute to itself. New commands and other tokens
or attributes will be added to that language.

Attributes:
    PATTERN (TYPE): The description of the type an argument (=pattern).
    VERBOSE (bool): Load plugins verbosely.
"""

from os import path
import re
from typing import Any, Callable, Dict, List, Optional, Type, Union

from pluginbase import PluginBase

import superspy
from superspy import ast


VERBOSE = True

PATTERN = Union[
    Type['Token'], List[Type['Token']]
]


def bounce(bounce_function: Callable) -> Callable:
    """Inner function required for some decorators. Takes a function
    and returns it.

    Args:
        bounce_function (Callable): The function to be returned.

    Returns:
        Callable: The function that has been passed to the function.
    """
    return bounce_function


##############
# DECORATORS #
##############


def register_data_type(regex: str) -> Any:
    """Registers a new data type matching a specified regex.

    Args:
        regex (str): The regular expression describing the data type.
    """

    def decorator_register(
            data_type_obj: Type['ast.DataType']) -> Type['ast.DataType']:
        """Inner function.

        Args:
            data_type_obj (Type['ast.DataType']): The class to be instantiated
                when matching the regex.

        Returns:
            Type['ast.DataType']: Required for the decorator.
        """
        Language.current.regex_matchers[regex] = data_type_obj
        return data_type_obj

    return decorator_register


def register_token_regex(regex: str) -> Any:
    """Registers a new token matching a specified regex.

    Args:
        regex (str): The regular expression describing the token.
    """

    def decorator_register(token_obj: Type['ast.Token']) -> Type['ast.Token']:
        """Inner function.

        Args:
            token_obj (Type['ast.Token']): The class to be instantiated
                when matching the regex.

        Returns:
            Type['ast.Token']: Required for the decorator.
        """
        Language.current.regex_matchers[regex] = token_obj
        return token_obj

    return decorator_register


def register_token(name: str) -> Any:
    """Register a new token matching a specified name.

    Args:
        name (str): The name matching the token.
    """
    # Escape the Regex, so for example `*` still works and does not match any
    # token. However newlines may not be escaped.
    escaped_re = name
    if name not in ['\n']:
        escaped_re = re.escape(name)

    return register_token_regex(f'^{escaped_re}$')


def register_word_delimiter(character: chr) -> Any:
    """Registers a new word delimiting character.

    Args:
        character (chr): The character that delimits words.
    """

    Language.current.word_delimiters.append(character)
    return bounce


# pylint: disable=invalid-name
# This name is required
def register_single_character_touching_token(character: chr) -> Any:
    """Register a single character token that may touch other characters.

    So for example in `6*7` the `*` is a separate token, even though it is
    touching other characters. Because writing `6 * 7` every time is
    too inconvenient in this world.

    Args:
        character (chr): The character that describes the token.
    """

    def decorator_register(token_obj: Type['ast.Token']) -> Type['ast.Token']:
        """Inner function.

        Args:
            token_obj (Type['ast.Token']): The class to be instantied
                when matching the character.

        Returns:
            Type['ast.Token']: Required for the decorator.
        """
        return register_token(character)(token_obj)

    Language.current.single_character_tokens.append(character)
    return decorator_register


def register_command_with_pattern(*pattern: List[PATTERN]) -> Any:
    """Register a command matching a pattern of arguments and tokens.

    Args:
        *pattern (List[PATTERN]): The pattern to be matched.
    """

    def decorator_register(
            command_obj: Type['ast.Command']) -> Type['ast.Command']:
        """Inner function.

        Args:
            command_obj (Type['ast.Command']): The class to be instantied
                when matching the pattern.

        Returns:
            Type['ast.Command']: Required for the decorator.
        """
        Language.current.all_commands.append(command_obj)

        priority = command_obj.priority
        if priority in Language.current.commands_by_priority:
            Language.current.commands_by_priority[priority].append(command_obj)
        else:
            Language.current.commands_by_priority[priority] = [command_obj]
        command_obj.pattern = pattern

        return command_obj

    return decorator_register


def register_command(command: Type['ast.Command']) -> Any:
    """Register a command matching the pattern described in the `__init__`.

    Args:
        command (Type['ast.Command']): The class to be instantiated
            when matching the character.
    """
    pattern = list(command.__init__.__annotations__.values())
    return register_command_with_pattern(*pattern)(command)


def register_literality_delimiter(opening: str, closing: str) -> Any:
    """Register a pair of strings delimiting literality, like "".

    So everything inside the delimiters is taken literal, even though it
    might contain things like spaces or even commands.

    Args:
        opening (str): The string beginning the literality.
        closing (str): The string closing   the literality.
    """

    Language.current.literality_delimiter[opening] = closing
    return bounce


def register_operator(operator: str) -> Any:
    """Register an operator like `*` or `==` or `equals`.

    Args:
        operator (str): The operator to be registered.
    """

    # pylint: disable=pointless-statement
    @register_single_character_touching_token(operator)
    class OperatorToken(ast.Token):
        """A dumb token representing the operator.
        """
        ...

    def decorator_register(
            function_obj: Type['ast.Function']) -> Type['ast.Function']:
        """Inner function.

        Args:
            function_obj (Type['ast.Function']): The class to be instantiated
                when matching the operator pattern.

        Returns:
            Type['ast.Function']: Required for the decorator.
        """
        pattern = list([ast.Command, OperatorToken, ast.Command])
        register_command_with_pattern(*pattern)(function_obj)
        return function_obj

    return decorator_register


def register_function(function_name: str, argument_count: int = 1) -> Any:
    """Register a function with a specified amount of arguments.

    Args:
        function_name (str): The name of the function.
        argument_count (int, optional): The count of arguments
    """

    # pylint: disable=pointless-statement
    @register_token(function_name)
    class FunctionNameToken(ast.Token):
        """A dumb token representing the function call.
        """
        ...

    def decorator_register(
            function_obj: Type['ast.Function']) -> Type['ast.Function']:
        """Inner function.

        Args:
            function_obj (Type['ast.Function']): The class ot be instantiated
                when matching the function.

        Returns:
            Type['ast.Function']: Required for the decorator.
        """
        pattern = list([FunctionNameToken, *([ast.Command] * argument_count)])
        register_command_with_pattern(*pattern)(function_obj)
        return function_obj

    return decorator_register


def register_falsy_value(falsy_value: str) -> Any:
    """Register a value that is falsy (or not True).

    So for example `while []` will not get executed.

    Args:
        falsy_value (str): The value that is falsy
    """

    Language.current.falsy_values.append(falsy_value)
    return bounce


###########
# CLASSES #
###########


class Language:
    """An object representing a language including commands and syntax.

    It is responsible for independently loading plugins from specified
    searchpaths, registering new commands, tokens etc. and storing this data
    including other attributes like literality delimiters or escape markers
    to be used by the abstract syntax tree for parsing.

    Attributes:
        excluded_plugins (List[str]): List of plugins to not be loaded in the
            specified searchpaths. So if you don't want `io` to be loaded, add
            it here.

        all_commands (List[Type['ast.Command']]): List of all registered
            commands.
        commands_by_priority (Dict[float, List[Type['ast.Command']]]): A
            dictionary that has the priorities (or order of operations) of all
            commands as keys and the command classes as values.

        escape_markers (str): Characters marking to next character to be
            escaped. So for example `\"` does not end the string.
        word_delimiters (List[str]): Characters delimiting words, like spaces.
        regex_matchers (Dict[str, Type['ast.Token']]): The regex to match
            tokens and the token to be instantiated.
        single_character_tokens (List[chr]): Tokens that are only one character
            long, like `*` or `=`.
        literality_delimiter (Dict[str, str]): Literality delimiting pair of
            strings, like `""`.
        falsy_values (List[str]): List of values that evaluate as false like
            `[]`, `""` or `0`.

        plugin_base (TYPE): PluginBase base object
        plugin_source (TYPE): PluginBase source object

        current (Language): Class variable for the currently loading language.
    """

    excluded_plugins: List[str] = []

    all_commands: List[Type['ast.Command']] = []
    commands_by_priority: Dict[float, List[Type['ast.Command']]] = {}
    escape_markers = '\\'
    word_delimiters: List[chr] = []
    regex_matchers: Dict[str, Type['ast.Token']] = {}
    single_character_tokens: List[chr] = []
    literality_delimiter: Dict[str, str] = {}
    falsy_values: List[str] = []

    plugin_base: Any
    plugin_source: Any

    # classvar
    current: 'Language'

    def __init__(self, search_paths: List[str]):
        """Initiate the Language and load Plugins.

        Args:
            search_paths (List[str]): The search path for the plugins.
        """
        Language.current = self
        self.plugin_base = PluginBase(package=f'superspy.system')
        self.plugin_source = self.plugin_base.make_plugin_source(
            searchpath=search_paths)
        for plugin in self.plugin_source.list_plugins():
            if plugin not in self.excluded_plugins:
                if VERBOSE:
                    print(f'\033[36mLoaded plugin\033[0m {plugin}')
                self.plugin_source.load_plugin(plugin)

    def evaluates_as_true(self, statement: str) -> bool:
        """Return this statement evaluated as True.

        Args:
            statement (str): The statement to be evaluated

        Returns:
            bool: The return statements, True if the statement evaluates as
                True, False otherwise.
        """
        return statement not in self.falsy_values


class SuperSpyLanguage(Language):
    """A language subclass automatically loading the superspy plugins.
    """

    superspy_search_paths = ['system']
    default_search_paths = [
        path.join(path.dirname(superspy.__file__), superspy_search_path)
        for superspy_search_path in superspy_search_paths]

    def __init__(self,
                 custom_search_paths: Optional[List[str]] = None,
                 excluded_plugins: Optional[List[str]] = None):
        """Summary

        Args:
            custom_search_paths (Optional[List[str]], optional): Adding custom
                searchpaths, for plugins used for this shell.
            excluded_plugins (Optional[List[str]], optional): Ignore specific
                plugins. These can be in the custom search path, or in the
                default library. So for example ['io', 'arithmetic'] would not
                load those two plugins.
        """
        search_paths = self.default_search_paths + (custom_search_paths or [])
        self.excluded_plugins = (excluded_plugins or [])
        super().__init__(search_paths)



"""Basic arithmetic/math functions within the superspy language.
"""

from superspy import ast, language


@language.register_operator('*')
class Multiplication(ast.Operation):
    """Multiply two numbers.
    """

    priority = ast.OrderOfOperations.MULTIPLICATION

    def execute(self) -> float:
        """Execute the multiplication.

        Returns:
            float: The product.
        """
        return self.left.execute() * self.right.execute()

    def __repr__(self) -> str:
        return f'MULTIPLICATION ({repr(self.left)}) TIMES ({repr(self.right)})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}MULTIPLICATION\n'\
            f'{self.left.deep_repr(indent+self.addint)}\n'\
            f'{indent+self.addint}*\n'\
            f'{self.right.deep_repr(indent+self.addint)}'


@language.register_operator('/')
class Division(ast.Operation):
    """Divide two numbers.
    """

    priority = ast.OrderOfOperations.MULTIPLICATION

    def execute(self) -> float:
        """Execute the Division.

        Returns:
            float: The ratio.
        """
        return self.left.execute() / self.right.execute()

    def __repr__(self) -> str:
        return f'DIVISION ({repr(self.left)}) OVER ({repr(self.right)})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}DIVISION\n'\
            f'{self.left.deep_repr(indent+self.addint)}\n'\
            f'{indent+self.addint}/\n'\
            f'{self.right.deep_repr(indent+self.addint)}'


@language.register_operator('+')
class Addition(ast.Operation):
    """Add two numbers.
    """

    priority = ast.OrderOfOperations.ADDITION

    def execute(self) -> float:
        """Executes the addition.

        Returns:
            float: The sum.
        """
        return self.left.execute() + self.right.execute()

    def __repr__(self) -> str:
        return f'ADDITION ({repr(self.left)}) PLUS ({repr(self.right)})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}ADDITION\n'\
            f'{self.left.deep_repr(indent+self.addint)}\n'\
            f'{indent+self.addint}+\n'\
            f'{self.right.deep_repr(indent+self.addint)}'


@language.register_operator('-')
class Subtraction(ast.Operation):
    """Subtract two numbers.
    """

    priority = ast.OrderOfOperations.ADDITION

    def execute(self) -> float:
        """Executes the subtraction.

        Returns:
            float: The dfference.
        """
        return self.left.execute() - self.right.execute()

    def __repr__(self) -> str:
        return f'SUBTRACTION ({repr(self.left)}) MINUS ({repr(self.right)})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}DIFFERENCE\n'\
            f'{self.left.deep_repr(indent+self.addint)}\n'\
            f'{indent+self.addint}-\n'\
            f'{self.right.deep_repr(indent+self.addint)}'


@language.register_operator('==')
class Equivalence(ast.Operation):
    """Checks for equivalence of two values.
    """

    priority = ast.OrderOfOperations.EQUAL_COMPARISON

    def execute(self) -> bool:
        """Executes the comparison.

        Returns:
            bool: The equivalence.
        """
        return self.left.execute() == self.right.execute()

    def __repr__(self) -> str:
        return f'EQUIVALENCE ({repr(self.left)}) EQUALS ({repr(self.right)})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}EQUIVALENCE\n'\
            f'{self.left.deep_repr(indent+self.addint)}\n'\
            f'{indent+self.addint}==\n'\
            f'{self.right.deep_repr(indent+self.addint)}'


@language.register_operator('!=')
class UnEquivalence(ast.Operation):
    """Checks for unequivalence (!=) of to values.
    """

    priority = ast.OrderOfOperations.EQUAL_COMPARISON

    def execute(self) -> bool:
        """Executes the comparison.

        Returns:
            bool: The unequivalence.
        """
        return self.left.execute() != self.right.execute()

    def __repr__(self) -> str:
        return f'UNEQUIVALENCE ({repr(self.left)}) IS NOT ({repr(self.right)})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}UNEQUIVALENCE\n'\
            f'{self.left.deep_repr(indent+self.addint)}\n'\
            f'{indent+self.addint}!=\n'\
            f'{self.right.deep_repr(indent+self.addint)}'


@language.register_function('not')
class Not(ast.Function):
    """Invertes a value.
    This also works for non boolean value as it will check for truthyness and
    invert that.
    """

    priority = ast.OrderOfOperations.EXPRESSION

    def execute(self) -> bool:
        """Execute the operation.

        Returns:
            bool: Not the input value.
        """
        return '0' if self.argument.executes_as_true() else '1'

    def __repr__(self) -> str:
        return f'NOT ({self.args[0]})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}NOT\n{self.args[0].deep_repr(indent+self.addint)}'



"""Basic language structure for the superspy language.
"""

from typing import Any

from superspy import ast, language

# Delimit words with spaces
@language.register_word_delimiter(' ')
# Delimit commands with semicolons and newlines
@language.register_single_character_touching_token(';')
@language.register_single_character_touching_token('\n')
class NewLine(ast.Token):
    """Newline token.
    """
    def deep_repr(self, indent='') -> str:
        return super().deep_repr(indent) + '\n'


@language.register_command
class Line(ast.Command):
    """A line containing a single command.

    Attributes:
        command (ast.Command): The command making up the line.
    """

    command: ast.Command
    priority = ast.OrderOfOperations.COMMA
    can_contain_self = False

    def __init__(self, command: ast.Command, _: NewLine):
        """Initialize a line object.

        Args:
            command (ast.Command): The command.
            _ (NewLine): The new line character token.
        """
        super().__init__()
        self.command = command

    def execute(self) -> Any:
        """Execute the line by executing its single command.

        Returns:
            Any: The return value of the command.
        """
        return self.command.execute()

    def get_trace(self):
        return f'LINE {self.line_number}'

    def __repr__(self) -> str:
        return f'LINE {self.line_number}: {repr(self.command)}\n'

    def deep_repr(self, indent='') -> str:
        return f'{indent}LINE[{self.line_number}]\n'\
            f'{self.command.deep_repr(indent+self.addint)}\n'


@language.register_command
class EmptyLine(ast.Command):
    """An empty line not containing any command.
    This is needed to, after all other lines have matched match the remaining
    newline tokens.
    """

    priority = ast.OrderOfOperations.COMMA + 0.1

    def __init__(self, _: NewLine):
        """Initialize the empty line.

        Args:
            _ (NewLine): The new line character token.
        """
        super().__init__()

    def execute(self):
        """Execute an empty line - do nothing.
        """
        pass

    def __repr__(self) -> str:
        return f'LINE {self.line_number}: EMPTY\n'

    def deep_repr(self, indent='') -> str:
        return f'{indent}LINE[{self.line_number}]: EMPTY\n'



"""Summary
"""

from superspy import ast, language


@language.register_token('getnum')
class GetNum(ast.Command):
    """Get a number from user input.
    """

    def execute(self) -> float:
        """Ask for input and convert it to a float.

        Returns:
            float: The user input.
        """
        return float(input())

    def __repr__(self) -> str:
        return f'GET NUMBER'

    def deep_repr(self, indent='') -> str:
        return f'{indent}GET NUMBER'


@language.register_token('getstr')
class GetStr(ast.Command):
    """Get a string from user input.
    """

    def execute(self) -> str:
        """Ask for input.

        Returns:
            str: The input.
        """
        return input()

    def __repr__(self) -> str:
        return f'GET STRING'

    def deep_repr(self, indent='') -> str:
        return f'{indent}GET STRING'


@language.register_function('print', 1)
class Print(ast.Function):
    """Print a given argument.
    """

    def execute(self):
        """Execute the argument and print it.
        """
        print(self.argument.execute())

    def __repr__(self) -> str:
        return f'PRINT ({self.argument})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}PRINT\n{self.argument.deep_repr(indent+self.addint)}'


@language.register_function('printnl', 1)
class PrintNL(ast.Function):
    """Print a given argument without a newline (No Line).
    """

    def execute(self):
        """Execute the argument and print it.
        """
        print(self.argument.execute(), end='')

    def __repr__(self) -> str:
        return f'PRINTNL ({self.argument})'

    def deep_repr(self, indent='') -> str:
        return f'{indent}PRINTNL\n'\
            f'{self.argument.deep_repr(indent+self.addint)}'


@language.register_token('dis')
class Disassemble(ast.Command):
    """Print the token tree built by the ast.
    """

    def execute(self):
        """Call the corresponding function in the ast.
        """
        print(self.ast.runtime.root_ast.deep_repr())

    def __repr__(self) -> str:
        return f'DISASSEMBLE'

    def deep_repr(self, indent='') -> str:
        return f'{indent}DISASSEMBLE'



from setuptools import setup

with open('readme.md') as file:
    long_description = file.read().split('<!---START--->')[-1].split('<!---END--->')[0]
    long_description = long_description.replace('<!---PYPI', '')
    long_description = long_description.replace('PYPI--->', '')

setup(
    name='superspy',
    version='0.3.4',
    description='Small Uncomplicated Plugin Extensible Reliable Shell in PYthon',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Kamik423/superspy',
    author='Hans',
    author_email='contact.kamik423@gmail.com',
    license='MIT',
    packages=['superspy', 'superspy.system'],
    zip_safe=False,
    install_requires=['pluginbase'],
    classifiers=(
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    )
)



"""Code Sources provide lines for the code.

The responsibility of a code source to provide an abstract syntax tree with
new lines while parsing and/or interpreting.
"""

from abc import ABC
from typing import List


class CodeSource(ABC):
    """This is the an abstract code source base.

    It has to be inherited from, but already provides a mechanism, that, if not
    overwritten will feed all of it's lines, in the code variable to the
    abstract syntax tree.

    Attributes:
        code (List[str]): A list of lines, each a string. Does not have to be
            used by an inheriting class, but provides the auto-feed-mechanism.
            Each line should end in `\n`.
        execution_index (int): The current index in the code variable that has
            to be returned next.
    """

    code: List[str] = ''
    execution_index: int = 0

    # pylint: disable=unused-argument
    # because is_valid is unused in this case, but supposed to be usable by
    # inheriting classes
    def next_line(self, is_valid: bool) -> str:
        """Return the next line.

        This method gets called when the abstract syntax tree requires another
        line of code.

        Args:
            is_valid (bool): Is the syntax tree currently valid. Used to
                provide an indent in the ShellSource. For example.
                Not used by the default CodeSource.

        Returns:
            str: The next line.
        """
        line = self.code[self.execution_index]
        self.execution_index += 1
        return line

    def has_more_lines(self) -> bool:
        """Does this code source have any more lines.

        Returns:
            bool: The return value. True if the code source can still provide
                more lines. False if it reached the end of it's source. The
                shell still returns True, even if it does not have any more
                lines ready to be returned, because it wants the abstract
                syntax tree to query for more lines.
        """
        return self.execution_index != len(self.code)


class StringSource(CodeSource):
    """A code source that takes its entire script from a string.
    """

    def __init__(self, code: str):
        """Initialize the code source and pass it its code string.

        Args:
            code (str): The string that the code source runs from.
        """
        self.code = [f'{line}\n' for line in code.split('\n')]


class FileSource(StringSource):
    """A code source that takes its entire script from a file.
    """

    def __init__(self, filename: str):
        """Initialize the code source and pass it its file.

        Args:
            filename (str): The name of the script (relative to the main
            Python script) to be executed
        """
        with open(filename, 'r') as f:
            super().__init__(f.read())


class ShellSource(CodeSource):
    """A code source that takes its script from an interactive shell using
    `input()`.

    Attributes:
        prompt (str): The prompt to be displayed at the front of each line in
            the shell. Should end if a space.
    """

    prompt = '[SUPERSPY] '

    def next_line(self, is_valid: bool) -> str:
        """Return the next line.

        Retrieved from `input()`. If the current abstract syntax tree is not
        valid it assumes that there currently is an unused closing brace and
        thus indents the line.

        Args:
            is_valid (bool): Is the syntax tree currently valid. Used to
                provide an indent.

        Returns:
            str: The next line.
        """
        indent = '' if is_valid else '    '
        return input(f'{self.prompt}{indent}')

    def has_more_lines(self) -> bool:
        """Does this code source have any more lines.

        Returns:
            bool: Always True, as the shell will always prompt for more lines
                unless the syntax tree interpreter terminates, at which point
                it will stop asking for lines.
        """
        return True
